from http.client import HTTPException

import gevent.server
import gevent.socket
import pytest

from geventhttpclient.client import CRLF, HTTPClient
from tests.common import LISTENER, server


def wrong_response_status_line(sock, addr):
    sock.recv(4096)
    sock.sendall(b"HTTP/1.1 apfais df0 asdf\r\n\r\n")


def test_exception():
    with server(wrong_response_status_line):
        connection = HTTPClient(*LISTENER)
        with pytest.raises(HTTPException):
            connection.get("/")


def close(sock, addr):
    sock.close()


def test_close():
    with server(close):
        client = HTTPClient(*LISTENER)
        with pytest.raises(HTTPException):
            client.get("/")


def close_after_recv(sock, addr):
    sock.recv(4096)
    sock.close()


def test_close_after_recv():
    with server(close_after_recv):
        client = HTTPClient(*LISTENER)
        with pytest.raises(HTTPException):
            client.get("/")


def timeout_recv(sock, addr):
    sock.recv(4096)
    gevent.sleep(1)


def test_timeout_recv():
    with server(timeout_recv):
        connection = HTTPClient(*LISTENER, network_timeout=0.1)
        with pytest.raises(gevent.socket.timeout):
            connection.request("GET", "/")


def timeout_send(sock, addr):
    gevent.sleep(1)


def test_timeout_send():
    with server(timeout_send):
        connection = HTTPClient(*LISTENER, network_timeout=0.1)
        with pytest.raises(gevent.socket.timeout):
            connection.request("GET", "/")


def close_during_content(sock, addr):
    sock.recv(4096)
    sock.sendall(b"""HTTP/1.1 200 Ok\r\nContent-Length: 100\r\n\r\n""")
    sock.close()


def test_close_during_content():
    with server(close_during_content):
        client = HTTPClient(*LISTENER, block_size=1)
        response = client.get("/")
        with pytest.raises(HTTPException):
            response.read()


def content_too_small(sock, addr):
    sock.recv(4096)
    sock.sendall(b"""HTTP/1.1 200 Ok\r\nContent-Length: 100\r\n\r\ncontent""")
    gevent.sleep(10)


def test_content_too_small():
    with server(content_too_small):
        client = HTTPClient(*LISTENER, network_timeout=0.2)
        with pytest.raises(gevent.socket.timeout):
            response = client.get("/")
            response.read()


def close_during_chunked_readline(sock, addr):
    sock.recv(4096)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nTransfer-Encoding: chunked\r\n\r\n")

    chunks = [
        "This is the data in the first chunk\r\n",
        "and this is the second one\r\n",
        "con\r\n",
    ]

    for chunk in chunks:
        gevent.sleep(0.1)
        sock.sendall((hex(len(chunk))[2:] + CRLF + chunk + CRLF).encode())
    sock.close()


def test_close_during_chunked_readline():
    with server(close_during_chunked_readline):
        client = HTTPClient(*LISTENER)
        response = client.get("/")
        assert response["transfer-encoding"] == "chunked"
        chunks = []
        with pytest.raises(HTTPException):
            data = "enter_loop"
            while data:
                data = response.readline()
                chunks.append(data)
        assert len(chunks) == 3


def timeout_during_chunked_readline(sock, addr):
    sock.recv(4096)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nTransfer-Encoding: chunked\r\n\r\n")

    chunks = [
        "This is the data in the first chunk\r\n",
        "and this is the second one\r\n",
        "con\r\n",
    ]

    for chunk in chunks:
        sock.sendall((hex(len(chunk))[2:] + CRLF + chunk + CRLF).encode())
    gevent.sleep(2)
    sock.close()


def test_timeout_during_chunked_readline():
    with server(timeout_during_chunked_readline):
        client = HTTPClient(*LISTENER, network_timeout=0.1)
        response = client.get("/")
        assert response["transfer-encoding"] == "chunked"
        chunks = []
        with pytest.raises(gevent.socket.timeout):
            data = "enter_loop"
            while data:
                data = response.readline()
                chunks.append(data)
        assert len(chunks) == 3



from collections.abc import Mapping
from urllib import parse as urlparse

DEFAULT_PORTS = {"http": 80, "https": 443}


class URL:
    """Immutable URL class

    You build it from an url string.
    >>> url = URL('http://python.org/urls?param=asdfa')
    >>> url
    URL(http://python.org/urls?param=asdfa)

    You cast it to a tuple, it returns the same tuple as `urlparse.urlsplit`.
    >>> tuple(url)
    ('http', 'python.org', '/urls', 'param=asdfa', '')

    You can cast it as a string.
    >>> str(url)
    'http://python.org/urls?param=asdfa'
    """

    __slots__ = ("_parsed",)

    def __init__(self, url="", params=None):
        if isinstance(url, str):
            parsed = urlparse.urlparse(url)
        else:
            parsed = url
        scheme, netloc, path, parsed_params, query, fragment = parsed

        if params is not None:
            new_params = _encode_params(params)
            query = query + "&" + new_params if query else new_params
        self._parsed = urlparse.ParseResult(scheme, netloc, path, parsed_params, query, fragment)

    def __str__(self):
        return self._parsed.geturl()

    def __repr__(self):
        return f"URL({self})"

    def __iter__(self):
        return (val if val is not None else "" for val in self._parsed)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            other = type(self)(other)
        return self._parsed == other._parsed

    def __getattr__(self, attr):
        value = getattr(self._parsed, attr)
        # backwards compatibility: never return None for URL parts
        return value if value is not None else ""

    @property
    def host(self):
        return self.hostname

    @property
    def user(self):
        return self.username

    @property
    def port(self):
        port = self._parsed._hostinfo[1]
        if port is not None:
            if port.isdigit() and port.isascii():
                port = int(port)
            else:
                raise ValueError(f"Port could not be cast to integer value as {port!r}")
            if not (0 <= port <= 65535):
                raise ValueError("Port out of range 0-65535")
        else:
            port = DEFAULT_PORTS.get(self._parsed.scheme)
        return port

    @property
    def query_string(self):
        return self.query

    @property
    def request_uri(self):
        if not self.query:
            return self.path
        return self.path + "?" + self.query

    def redirect(self, other):
        """Redirect to the other URL, relative to the current one."""
        if isinstance(other, str):
            other = URL(other)

        if other.netloc:
            return other

        # relative redirect
        scheme, netloc, path, params, query, fragment = other
        scheme = self.scheme
        netloc = self.netloc
        if not path.startswith("/"):
            if path.endswith("/"):
                path = self.path + path
            else:
                path = self.path.rstrip("/") + "/" + path
        parsed = urlparse.ParseResult(scheme, netloc, path, params, query, fragment)
        return type(self)(parsed)

    @property
    def quoted(self):
        return requote_uri(str(self))

    @property
    def quoted_uri(self):
        return requote_uri(self.request_uri)


def _encode_params(data):
    """Encode parameters in a piece of data.
    Will successfully encode parameters when passed as a dict or a list of 2-tuples.
    """

    if isinstance(data, (str, bytes)):
        return data
    if hasattr(data, "__iter__"):
        result = []
        for k, vs in to_key_val_list(data):
            if isinstance(vs, (str, bytes)) or not hasattr(vs, "__iter__"):
                vs = [vs]
            for v in vs:
                if v is not None:
                    result.append(
                        (
                            k.encode("utf-8") if isinstance(k, str) else k,
                            v.encode("utf-8") if isinstance(v, str) else v,
                        )
                    )
        return urlparse.urlencode(result, doseq=True)
    return data


def to_key_val_list(value):
    """Take an object and test to see if it can be represented as a
    dictionary. If it can be, return a list of tuples, e.g.,
    ::
        >>> to_key_val_list([('key', 'val')])
        [('key', 'val')]
        >>> to_key_val_list({'key': 'val'})
        [('key', 'val')]
        >>> to_key_val_list('string')
        Traceback (most recent call last):
        ...
        ValueError: cannot encode objects that are not 2-tuples
    :rtype: list
    """
    if value is None:
        return None

    if isinstance(value, (str, bytes, bool, int)):
        raise ValueError("cannot encode objects that are not 2-tuples")

    if isinstance(value, Mapping):
        value = value.items()

    return list(value)


class InvalidURL(Exception):
    pass


# The following functions are taken from requests
# Copyright of the original requests project:
# :copyright: (c) 2012 by Kenneth Reitz.
# :license: Apache2, see LICENSE for more details.

# The unreserved URI characters (RFC 3986)
UNRESERVED_SET = frozenset(
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz" + "0123456789-._~"
)


def unquote_unreserved(uri):
    """Un-escape any percent-escape sequences in a URI that are unreserved
    characters. This leaves all reserved, illegal and non-ASCII bytes encoded.

    :rtype: str
    """
    parts = uri.split("%")
    for i in range(1, len(parts)):
        h = parts[i][0:2]
        if len(h) == 2 and h.isalnum():
            try:
                c = chr(int(h, 16))
            except ValueError:
                raise InvalidURL(f"Invalid percent-escape sequence: '{h}'")

            if c in UNRESERVED_SET:
                parts[i] = c + parts[i][2:]
            else:
                parts[i] = f"%{parts[i]}"
        else:
            parts[i] = f"%{parts[i]}"
    return "".join(parts)


def requote_uri(uri):
    """Re-quote the given URI.

    This function passes the given URI through an unquote/quote cycle to
    ensure that it is fully and consistently quoted.

    :rtype: str
    """
    safe_with_percent = "!#$%&'()*+,/:;=?@[]~"
    safe_without_percent = "!#$&'()*+,/:;=?@[]~"
    try:
        # Unquote only the unreserved characters
        # Then quote only illegal characters (do not quote reserved,
        # unreserved, or '%')
        return urlparse.quote(unquote_unreserved(uri), safe=safe_with_percent)
    except InvalidURL:
        # We couldn't unquote the given URI, so let's try quoting it, but
        # there may be unquoted '%'s in the URI. We need to make sure they're
        # properly quoted so they do not cause issues elsewhere.
        return urlparse.quote(uri, safe=safe_without_percent)



import gevent.pool
import pytest

import geventhttpclient.httplib

httplib2 = pytest.importorskip("geventhttpclient.httplib2")


def job(client, url):
    response, content = client.request(url)
    assert content
    assert b"body" in content


@pytest.mark.network
def test_request_parallel():
    with geventhttpclient.httplib.patched():
        errors = []

        client = httplib2.Http(concurrency=5)
        group = gevent.pool.Pool(size=5)

        urls = ["https://google.com", "http://gevent.org", "https://github.com"]
        for url in urls:
            g = group.spawn(job, client, url)
            g.link_exception(lambda g: errors.append(g.exception))
        group.join()
        if errors:
            for e in errors:
                print(e)
            raise errors[0]



import pytest

from geventhttpclient.header import Headers
from geventhttpclient.requests import Session
from tests.common import HTTPBIN_HOST


@pytest.mark.network
def test_no_form_encode_header():
    url = f"https://{HTTPBIN_HOST}/headers"
    hdrs = Headers(Session().get(url).json()["headers"])
    print(hdrs)
    assert "content-type" not in hdrs
    assert "content-length" not in hdrs



import json
import socket

import gevent.pool
import gevent.queue
import gevent.server
import pytest

from geventhttpclient import __version__
from geventhttpclient.client import METHOD_GET, HTTPClient
from geventhttpclient.connectionpool import ConnectionPool
from tests.common import HTTPBIN_HOST, LISTENER, check_upload, server, wsgiserver


def httpbin_client(
    host=HTTPBIN_HOST,
    port=None,
    headers=None,
    block_size=HTTPClient.BLOCK_SIZE,
    connection_timeout=30.0,
    network_timeout=30.0,
    disable_ipv6=True,
    concurrency=1,
    ssl=False,
    ssl_options=None,
    ssl_context_factory=None,
    insecure=False,
    proxy_host=None,
    proxy_port=None,
    version=HTTPClient.HTTP_11,
):
    """Client factory for httpbin with higher timeout values"""

    return HTTPClient(
        host,
        port=port,
        headers=headers,
        block_size=block_size,
        connection_timeout=connection_timeout,
        network_timeout=network_timeout,
        disable_ipv6=disable_ipv6,
        concurrency=concurrency,
        ssl=ssl,
        ssl_options=ssl_options,
        ssl_context_factory=ssl_context_factory,
        insecure=insecure,
        proxy_host=proxy_host,
        proxy_port=proxy_port,
        version=version,
    )


@pytest.fixture
def httpbin():
    return httpbin_client()


@pytest.mark.parametrize("request_uri", ["/tp/", "tp/", f"http://{HTTPBIN_HOST}/tp/"])
def test_build_request(httpbin, request_uri):
    request_ref = f"GET /tp/ HTTP/1.1\r\nUser-Agent: python/gevent-http-client-{__version__}\r\nHost: {HTTPBIN_HOST}\r\n\r\n"
    assert httpbin._build_request(METHOD_GET, request_uri) == request_ref


def test_build_request_invalid_host(httpbin):
    with pytest.raises(ValueError):
        httpbin._build_request(METHOD_GET, "http://someunrelatedhost.com/")


@pytest.mark.parametrize("port", [None, 1234])
@pytest.mark.parametrize("host", ["localhost", "127.0.0.1", "::1", "[::1]"])
def test_build_request_host(host, port):
    client = HTTPClient(host, port)
    host_ref = (
        f"host: {f'[{host}]' if host.startswith(':') else host}{f':{port}' if port else ''}\r\n"
    )
    assert host_ref in client._build_request(METHOD_GET, "").lower()


test_url_client_args = [
    ("http://python.org", ("python.org", 80)),
    ("http://python.org:333", ("python.org", 333)),
]


@pytest.mark.parametrize(["url", "client_args"], test_url_client_args)
def test_from_url(url, client_args):
    from_url = HTTPClient.from_url(url)
    from_args = HTTPClient(*client_args)
    assert from_args.host == from_url.host
    assert from_args.port == from_url.port


class StreamTestIterator:
    def __init__(self, sep, count):
        lines = [json.dumps({"index": i, "title": f"this is line {i}"}) for i in range(0, count)]
        self.buf = (sep.join(lines) + sep).encode()

    def __len__(self):
        return len(self.buf)

    def __iter__(self):
        self.cursor = 0
        return self

    def next(self):
        if self.cursor >= len(self.buf):
            raise StopIteration()

        gevent.sleep(0)
        pos = self.cursor + 10
        data = self.buf[self.cursor : pos]
        self.cursor = pos

        return data

    def __next__(self):
        return self.next()


def readline_iter(sock, addr):
    sock.recv(1024)
    iterator = StreamTestIterator("\n", 100)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nConnection: close\r\n\r\n")
    for block in iterator:
        sock.sendall(block)


def test_readline():
    with server(readline_iter):
        client = HTTPClient(*LISTENER, block_size=1)
        response = client.get("/")
        lines = []
        while True:
            line = response.readline(b"\n")
            if not line:
                break
            data = json.loads(line[:-1].decode())
            lines.append(data)
        assert len(lines) == 100
        assert [x["index"] for x in lines] == [x for x in range(0, 100)]


def chunks_iter(sock, addr):
    sock.recv(1024)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nContent-Length: 10\r\nConnection: close\r\n\r\n0123456789")


def test_response_chunks_iter():
    with server(chunks_iter):
        client = HTTPClient(*LISTENER, block_size=4)
        response = client.get("/")
        chunks = [next(response)]
        for chunk in response:
            chunks.append(chunk)
        assert b"".join(chunks) == b"0123456789"


def readline_multibyte_sep(sock, addr):
    sock.recv(1024)
    iterator = StreamTestIterator("\r\n", 100)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nConnection: close\r\n\r\n")
    for block in iterator:
        sock.sendall(block)


def test_readline_multibyte_sep():
    with server(readline_multibyte_sep):
        client = HTTPClient(*LISTENER, block_size=1)
        response = client.get("/")
        lines = []
        while True:
            line = response.readline(b"\r\n")
            if not line:
                break
            data = json.loads(line[:-1].decode())
            lines.append(data)
        assert len(lines) == 100
        assert [x["index"] for x in lines] == [x for x in range(0, 100)]


def readline_multibyte_splitsep(sock, addr):
    sock.recv(1024)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nConnection: close\r\n\r\n")
    sock.sendall(b'{"a": 1}\r')
    gevent.sleep(0)
    sock.sendall(b'\n{"a": 2}\r\n{"a": 3}\r\n')


def test_readline_multibyte_splitsep():
    with server(readline_multibyte_splitsep):
        client = HTTPClient(*LISTENER, block_size=1)
        response = client.get("/")
        lines = []
        last_index = 0
        while True:
            line = response.readline(b"\r\n")
            if not line:
                break
            data = json.loads(line[:-2].decode())
            assert data["a"] == last_index + 1
            last_index = data["a"]
        len(lines) == 3


def internal_server_error(sock, addr):
    sock.recv(1024)
    head = (
        "HTTP/1.1 500 Internal Server Error\r\n"
        "Connection: close\r\n"
        "Content-Type: text/html\r\n"
        "Content-Length: 135\r\n\r\n"
    )

    body = (
        "<html>\n  <head>\n    <title>Internal Server Error</title>\n  "
        "</head>\n  <body>\n    <h1>Internal Server Error</h1>\n    \n  "
        "</body>\n</html>\n\n"
    )

    sock.sendall((head + body).encode())
    sock.close()


def test_internal_server_error():
    with server(internal_server_error):
        client = HTTPClient(*LISTENER)
        response = client.get("/")
        assert not response.should_keep_alive()
        assert response.should_close()
        body = response.read()
        assert len(body) == response.content_length


def test_file_post(tmp_path):
    fpath = tmp_path / "tmp_body.txt"
    with open(fpath, "wb") as body:
        body.write(b"123456789")
    with wsgiserver(check_upload(b"123456789", length=9)):
        client = HTTPClient(*LISTENER)
        with open(fpath, "rb") as body:
            client.post("/", body)


def test_bytes_post():
    with wsgiserver(check_upload(b"12345", length=5)):
        client = HTTPClient(*LISTENER)
        client.post("/", b"12345")


def test_string_post():
    with wsgiserver(check_upload(b"12345", length=5)):
        client = HTTPClient(*LISTENER)
        client.post("/", "12345")


def test_unicode_post():
    byte_string = b"\xc8\xb9\xc8\xbc\xc9\x85"
    unicode_string = byte_string.decode("utf-8")
    with wsgiserver(check_upload(byte_string, length=len(byte_string))):
        client = HTTPClient(*LISTENER)
        client.post("/", unicode_string)


# The tests below require online access. We should try to replace them at least
# partly with local testing solutions and have the online tests as an extra on top.


@pytest.mark.network
def test_client_simple(httpbin):
    assert httpbin.port == 80
    response = httpbin.get("/")
    assert response.status_code == 200
    body = response.read()
    assert len(body)


@pytest.mark.network
def test_client_without_leading_slash(httpbin):
    with httpbin.get("") as response:
        assert response.status_code == 200
    with httpbin.get("base64/test") as response:
        assert response.status_code in (200, 301, 302)


FIREFOX_USER_AGENT = (
    "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:124.0) Gecko/20100101 Firefox/124.0"
)
FIREFOX_HEADERS = {"User-Agent": FIREFOX_USER_AGENT}


def check_user_agent_header(ua_header, ua_header_ref):
    """
    Unlike original httpbin, httpbingo.org sends back a list of header
    strings instead of a simple string. So we need to be a bit flexible
    with the answer.
    """
    if isinstance(ua_header, list):
        assert len(ua_header) == 1
        assert ua_header[0] == ua_header_ref
        return
    assert ua_header == ua_header_ref


@pytest.mark.network
def test_client_with_default_headers():
    httpbin = httpbin_client(headers=FIREFOX_HEADERS)
    response = httpbin.get("/headers")
    assert response.status_code == 200
    sent_headers = json.loads(response.read().decode())["headers"]
    check_user_agent_header(sent_headers["User-Agent"], FIREFOX_USER_AGENT)


@pytest.mark.network
def test_request_with_headers(httpbin):
    response = httpbin.get("/headers", headers=FIREFOX_HEADERS)
    assert response.status_code == 200
    sent_headers = json.loads(response.read().decode())["headers"]
    check_user_agent_header(sent_headers["User-Agent"], FIREFOX_USER_AGENT)


@pytest.mark.network
def test_response_context_manager(httpbin):
    r = None
    with httpbin.get("/") as response:
        assert response.status_code == 200
        r = response
    assert r._sock is None  # released


@pytest.mark.network
def test_multi_queries_greenlet_safe():
    httpbin = httpbin_client(concurrency=3)
    group = gevent.pool.Group()
    event = gevent.event.Event()

    def run(i):
        event.wait()
        response = httpbin.get("/")
        return response, response.read()

    count = 0
    ok_count = 0

    gevent.spawn_later(0.2, event.set)
    for response, content in group.imap_unordered(run, range(5)):
        # occasionally remotely hosted httpbin does return server errors :-/
        assert response.status_code in {200, 502, 504}
        if response.status_code == 200:
            ok_count += 1
        assert len(content)
        count += 1
    assert count == 5
    # ensure at least 3 of requests got 200
    assert ok_count >= 3


class TestIsSocketAlive:
    """Unit tests for ConnectionPool._is_socket_alive()"""

    def test_closed_socket_returns_false(self):
        sock = socket.socket()
        sock.close()
        pool = ConnectionPool("127.0.0.1", 80, "127.0.0.1", 80)
        assert sock.fileno() < 0
        assert pool._is_socket_alive(sock) is False

    def test_healthy_socket_returns_true(self):
        s1, s2 = socket.socketpair()
        try:
            pool = ConnectionPool("127.0.0.1", 80, "127.0.0.1", 80)
            assert pool._is_socket_alive(s1) is True
        finally:
            s1.close()
            s2.close()

    def test_peer_orderly_close_returns_false(self):
        s1, s2 = socket.socketpair()
        try:
            s2.close()
            pool = ConnectionPool("127.0.0.1", 80, "127.0.0.1", 80)
            assert pool._is_socket_alive(s1) is False
        finally:
            s1.close()



import pytest

from geventhttpclient.url import URL

url_full = "http://gevent.org/subdir/file.py?param=value&other=true#frag"
url_path_only = "/path/to/something?param=value&other=true"


def test_simple_url():
    url = URL(url_full)
    assert url.path == "/subdir/file.py"
    assert url.host == "gevent.org"
    assert url.port == 80
    assert url.query == "param=value&other=true"
    assert url.fragment == "frag"


def test_path_only():
    url = URL(url_path_only)
    assert url.host == ""
    assert url.port is None
    assert url.path == "/path/to/something"
    assert url.query == "param=value&other=true"


def test_params():
    url = URL(url_full, params={"pp": "hello"})
    assert url.path == "/subdir/file.py"
    assert url.host == "gevent.org"
    assert url.port == 80
    assert url.query == "param=value&other=true&pp=hello"
    assert url.fragment == "frag"


def test_params_urlencoded():
    url = URL(url_full, params={"a/b": "c/d"})
    assert url.path == "/subdir/file.py"
    assert url.host == "gevent.org"
    assert url.port == 80
    assert url.query == "param=value&other=true&a%2Fb=c%2Fd"
    assert url.fragment == "frag"


def test_quote_spaces():
    url = URL("http://gevent.org/?foo=bar with spaces")
    assert url.quoted == "http://gevent.org/?foo=bar%20with%20spaces"
    assert url.quoted_uri == "/?foo=bar%20with%20spaces"
    assert url.host == "gevent.org"
    assert url.port == 80


def test_quote_non_ascii():
    url = URL("http://127.0.0.1:8000/ы")
    assert url.quoted == "http://127.0.0.1:8000/%D1%8B"
    assert url.quoted_uri == "/%D1%8B"
    assert url.host == "127.0.0.1"
    assert url.port == 8000


def test_tuple_unpack():
    url = URL("http://gevent.org/somepath?foo=bar#frag")
    assert len(tuple(url)) == 6
    scheme, netloc, path, params, query, fragment = url
    assert scheme == "http"
    assert netloc == "gevent.org"
    assert path == "/somepath"
    assert query == "foo=bar"
    assert fragment == "frag"


def test_tuple_unpack_no_none():
    url = URL("http://gevent.org/")
    assert len(tuple(url)) == 6
    assert not any(val is None for val in tuple(url))


def test_empty():
    url = URL()
    assert url.host == ""
    assert not url.port
    assert url.query == ""
    assert url.fragment == ""
    assert url.netloc == ""
    assert str(url) == ""


def test_empty_path():
    assert URL("http://gevent.org").path == ""


def test_consistent_reparsing():
    for surl in (url_full, url_path_only):
        url = URL(surl)
        reparsed = URL(str(url))
        for attr in URL.__slots__:
            assert getattr(reparsed, attr) == getattr(url, attr)


def test_redirection_abs_path():
    url = URL(url_full)
    updated = url.redirect("/test.html")
    assert updated.host == url.host
    assert updated.port == url.port
    assert updated.path == "/test.html"
    assert updated.query == ""
    assert updated.fragment == ""


@pytest.mark.parametrize("redirection", ("test.html?key=val", "folder/test.html?key=val"))
def test_redirection_rel_path(redirection):
    url = URL(url_full)
    updated = url.redirect(redirection)
    assert updated.host == url.host
    assert updated.port == url.port
    assert updated.path.startswith("/subdir/")
    assert updated.path.endswith(redirection.split("?", 1)[0])
    assert updated.query == "key=val"
    assert updated.fragment == ""


def test_redirection_full_path():
    url_full2_plain = "http://google.de/index"
    url = URL(url_full)
    updated = url.redirect(url_full2_plain)
    url_full2 = URL(url_full2_plain)
    for attr in URL.__slots__:
        assert getattr(updated, attr) == getattr(url_full2, attr)
    assert str(url_full2) == url_full2_plain


def test_query():
    assert URL("/some/url", params={"a": "b", "c": 2}).query == "a=b&c=2"


def test_equality():
    assert URL("https://example.com/") != URL("http://example.com/")
    assert URL("http://example.com/") == URL("http://example.com/")


def test_default_port():
    assert URL("https://python.org").port == 443
    assert URL("http://gevent.org").port == 80
    assert URL("example.com").port is None


def test_pw():
    url = URL("http://asdf:dd@example.com/index.php?aaaa=bbbbb")
    assert url.host == "example.com"
    assert url.port == 80
    assert url.user == "asdf"
    assert url.password == "dd"


def test_pw_with_port():
    url = URL("http://asdf:dd@example.com:90/index.php?aaaa=bbbbb")
    assert url.host == "example.com"
    assert url.port == 90
    assert url.user == "asdf"
    assert url.password == "dd"


def test_ipv6():
    url = URL("http://[2001:db8:85a3:8d3:1319:8a2e:370:7348]/")
    assert url.host == "2001:db8:85a3:8d3:1319:8a2e:370:7348"
    assert url.port == 80
    assert url.user == ""


def test_ipv6_with_port():
    url = URL("https://[2001:db8:85a3:8d3:1319:8a2e:370:7348]:8080/")
    assert url.host == "2001:db8:85a3:8d3:1319:8a2e:370:7348"
    assert url.port == 8080
    assert url.user == ""



import sys
from functools import wraps
from http.client import HTTPException
from io import StringIO

import pytest

from geventhttpclient.response import HTTPResponse

RESPONSE = (
    "HTTP/1.1 301 Moved Permanently\r\nLocation: http://www.google.fr/\r\n"
    "Content-Type: text/html; charset=UTF-8\r\n"
    "Date: Thu, 13 Oct 2011 15:03:12 GMT\r\n"
    "Expires: Sat, 12 Nov 2011 15:03:12 GMT\r\n"
    "Cache-Control: public, max-age=2592000\r\n"
    "Server: gws\r\nContent-Length: 218\r\n"
    "X-XSS-Protection: 1; mode=block\r\n\r\n"
    '<HTML><HEAD><meta http-equiv="content-type" content="text/html;charset=utf-8">\n'
    "<TITLE>301 Moved</TITLE></HEAD><BODY>\n"
    '<H1>301 Moved</H1>\nThe document has moved\n<A HREF="http://www.google.fr/">here</A>.\r\n'
    "</BODY></HTML>\r\n"
)

# borrowed from gevent
# sys.gettotalrefcount is available only with python built with debug flag on
gettotalrefcount = getattr(sys, "gettotalrefcount", None)


def wrap_refcount(method):
    if gettotalrefcount is None:
        return method

    @wraps(method)
    def wrapped(*args, **kwargs):
        import gc

        gc.disable()
        gc.collect()
        deltas = []
        d = None
        try:
            for _ in range(4):
                d = gettotalrefcount()
                method(*args, **kwargs)
                if "urlparse" in sys.modules:
                    sys.modules["urlparse"].clear_cache()
                d = gettotalrefcount() - d
                deltas.append(d)
                if deltas[-1] == 0:
                    break
            else:
                raise AssertionError(f"refcount increased by {deltas!r}")
        finally:
            gc.collect()
            gc.enable()

    return wrapped


@wrap_refcount
def test_parse():
    parser = HTTPResponse()
    parser.feed(RESPONSE)
    assert parser.message_begun
    assert parser.headers_complete
    assert parser.message_complete


@wrap_refcount
def test_parse_small_blocks():
    parser = HTTPResponse()
    parser.feed(RESPONSE)
    response = StringIO(RESPONSE)
    while not parser.message_complete:
        data = response.read(10)
        parser.feed(data)

    assert parser.message_begun
    assert parser.headers_complete
    assert parser.message_complete
    assert parser.should_keep_alive()
    assert parser.status_code == 301
    assert sorted(parser.items()) == [
        ("Cache-Control", "public, max-age=2592000"),
        ("Content-Length", "218"),
        ("Content-Type", "text/html; charset=UTF-8"),
        ("Date", "Thu, 13 Oct 2011 15:03:12 GMT"),
        ("Expires", "Sat, 12 Nov 2011 15:03:12 GMT"),
        ("Location", "http://www.google.fr/"),
        ("Server", "gws"),
        ("X-XSS-Protection", "1; mode=block"),
    ]


@wrap_refcount
def test_parse_error():
    response = HTTPResponse()
    try:
        response.feed("HTTP/1.1 asdf\r\n\r\n")
        response.feed("")
        assert response.status_code, 0
        assert response.message_begun
    except HTTPException as e:
        assert "Invalid response status" in str(e)
    else:
        assert False, "should have raised"


@wrap_refcount
def test_incomplete_response():
    response = HTTPResponse()
    response.feed("""HTTP/1.1 200 Ok\r\nContent-Length:10\r\n\r\n1""")
    with pytest.raises(HTTPException):
        response.feed("")
    assert response.should_keep_alive()
    assert response.should_close()


@wrap_refcount
def test_response_too_long():
    response = HTTPResponse()
    data = """HTTP/1.1 200 Ok\r\nContent-Length:1\r\n\r\ntoolong"""
    with pytest.raises(HTTPException):
        response.feed(data)


@wrap_refcount
def test_on_body_raises():
    response = HTTPResponse()

    def on_body(buf):
        raise RuntimeError("error")

    response._on_body = on_body
    with pytest.raises(RuntimeError):
        response.feed(RESPONSE)


@wrap_refcount
def test_on_message_begin():
    response = HTTPResponse()

    def on_message_begin():
        raise RuntimeError("error")

    response._on_message_begin = on_message_begin
    with pytest.raises(RuntimeError):
        response.feed(RESPONSE)



import random
import string
from datetime import datetime
from email.message import Message
from http.cookiejar import CookieJar
from urllib.request import Request

import pytest

from geventhttpclient.header import Headers
from geventhttpclient.response import HTTPResponse

CUR_YEAR = datetime.now().year
LAST_YEAR = CUR_YEAR - 1
NEXT_YEAR = CUR_YEAR + 1
MULTI_COOKIE_RESPONSE = f"""
HTTP/1.1 200 OK
Server: nginx
Date: Fri, 21 Sep {CUR_YEAR} 18:49:35 GMT
Content-Type: text/html; charset=windows-1251
Connection: keep-alive
X-Powered-By: PHP/5.2.17
Set-Cookie: bb_lastvisit=1348253375; expires=Sat, 21-Sep-{NEXT_YEAR} 18:49:35 GMT; path=/
Set-Cookie: bb_lastactivity=0; expires=Sat, 21-Sep-{NEXT_YEAR} 18:49:35 GMT; path=/
Cache-Control: private
Pragma: private
Set-Cookie: bb_sessionhash=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_referrerid=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_userid=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_password=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_lastvisit=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_lastactivity=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_threadedmode=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_userstyleid=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_languageid=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_fbaccesstoken=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_fbprofilepicurl=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=/
Set-Cookie: bb_sessionhash=abcabcabcabcabcabcabcabcabcabcab; path=/; HttpOnly
Set-Cookie: tapatalk_redirect3=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_sessionhash=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: __utma=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: __utmb=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: __utmc=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: __utmz=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: vbulletin_collapse=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_referrerid=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_userid=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_password=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_lastvisit=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_lastactivity=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_threadedmode=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_userstyleid=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_languageid=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_fbaccesstoken=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Set-Cookie: bb_fbprofilepicurl=deleted; expires=Thu, 22-Sep-{LAST_YEAR} 18:49:34 GMT; path=1; domain=forum.somewhere.com
Content-Encoding: gzip
Content-Length: 26186

""".lstrip().replace("\n", "\r\n")
# Do not remove the final empty line!


def test_create_from_kwargs():
    h = Headers(ab=1, cd=2, ef=3, gh=4)
    assert len(h) == 4
    assert "ab" in h


def test_create_from_iterator():
    h = Headers((x, x * 5) for x in string.ascii_lowercase)
    assert len(h) == len(string.ascii_lowercase)


def test_create_from_dict():
    h = Headers(dict(ab=1, cd=2, ef=3, gh=4))
    assert len(h) == 4
    assert "ab" in h


def test_create_from_list():
    h = Headers([("ab", "A"), ("cd", "B"), ("cookie", "C"), ("cookie", "D"), ("cookie", "E")])
    assert len(h) == 5
    assert "ab" in h
    assert len(h["cookie"]) == 3
    assert h["cookie"][0] == "C"
    assert h["cookie"][-1] == "E"


def test_retrieve():
    h = Headers([("ab", "A"), ("cd", "B"), ("cookie", "C"), ("cookie", "D"), ("cookie", "E")])
    for key, ref in {"ab": "A", "cd": "B", "cookie": ["C", "D", "E"]}.items():
        assert h[key] == ref
        assert h.get(key) == ref
        assert h.pop(key) == ref
        assert key not in h


def test_case_insensitivity():
    h = Headers({"Content-Type": "text/plain"})
    h.add("Content-Encoding", "utf8")
    for val in ("content-type", "content-encoding"):
        assert val.upper() in h
        assert val.lower() in h
        assert val.capitalize() in h
        assert h.get(val.lower()) == h.get(val.upper()) == h.get(val.capitalize())
        del h[val.upper()]
        assert val.lower() not in h


def test_preserve_case():
    h = Headers(Cookie="C", COOKIE="D", cookie="E", asdf="E")
    assert len(h) == 4
    assert h["cookie"] == ["C", "D", "E"]
    assert list(h.items()) == [("Cookie", "C"), ("COOKIE", "D"), ("cookie", "E"), ("asdf", "E")]


def test_update_preserve_case():
    h = Headers()
    h.update(COOKIE="A", Cookie="C")
    assert list(h.items()) == [("Cookie", "C")]
    h = Headers()
    h.update(dict(Cookie="C", COOKIE="D", cookiE="E"))
    assert list(h.items()) == [("cookiE", "E")]


def test_read_multiple_header():
    parser = HTTPResponse()
    parser.feed(MULTI_COOKIE_RESPONSE)
    headers = parser._headers_index
    assert len(headers["set-cookie"]) == MULTI_COOKIE_RESPONSE.count("Set-Cookie")
    assert headers["set-cookie"][0].startswith("bb_lastvisit")
    assert headers["set-cookie"][-1].startswith("bb_fbprofilepicurl")


def test_cookielib_compatibility():
    cj = CookieJar()
    request = Request("https://forum.somewhere.com")
    parser = HTTPResponse()
    parser.feed(MULTI_COOKIE_RESPONSE)
    valid_cookie_count = sum(
        1
        for line in MULTI_COOKIE_RESPONSE.splitlines()
        if line.startswith("Set-Cookie") and str(LAST_YEAR) not in line
    )
    valid_cookies = cj.make_cookies(parser, request)
    assert len(valid_cookies) == valid_cookie_count
    cj.extract_cookies(parser, request)
    assert len(list(cj)) == valid_cookie_count


def test_compatibility_with_previous_api_read():
    parser = HTTPResponse()
    parser.feed(MULTI_COOKIE_RESPONSE)
    for single_item in (
        "content-encoding",
        "content-type",
        "content-length",
        "cache-control",
        "connection",
    ):
        assert isinstance(parser[single_item], str)
        assert isinstance(parser.get(single_item), str)


def test_compatibility_with_previous_api_write():
    h = Headers()
    h["asdf"] = "jklm"
    h["asdf"] = "dfdf"
    # Lists only if necessary
    assert h["asdf"] == "dfdf"


def test_copy():
    def rnd_txt(length):
        return "".join(random.choice(string.ascii_letters) for _ in range(length))

    h = Headers((rnd_txt(10), rnd_txt(50)) for _ in range(100))
    c = h.copy()
    assert h is not c
    assert len(h) == len(c)
    assert set(h.keys()) == set(c.keys())
    assert h == c
    assert type(h) is type(c)
    for _ in range(100):
        rnd_key = rnd_txt(9)
        c[rnd_key] = rnd_txt(10)
        assert rnd_key in c
        assert rnd_key not in h


def test_fieldname_string_enforcement():
    with pytest.raises(Exception):
        Headers({3: 3})
    h = Headers()
    with pytest.raises(Exception):
        h[3] = 5
    with pytest.raises(Exception):
        h.add(3, 4)
    with pytest.raises(Exception):
        del h[3]


def test_header_replace():
    d = Headers()
    d["Content-Type"] = "text/plain"
    d["content-type"] = "text/html"
    assert d["content-type"] == "text/html"


def test_formatting():
    h = Headers(asdf="ddd", ASDF="fff", AsDf="asdfasdf")
    assert str(h) == "asdf: ddd\nASDF: fff\nAsDf: asdfasdf"


def test_compat_dict():
    h = Headers(D="asdf")
    h.add("E", "d")
    h.add("E", "f")
    h.add("Cookie", "d")
    h.add("Cookie", "e")
    h.add("Cookie", "f")
    d = h.compatible_dict()

    for x in ("Cookie", "D", "E"):
        assert x in d
    assert d["D"] == "asdf"
    assert d["E"] == "d, f"
    assert d["Cookie"] == "d, e, f"


def test_extend_with_keys():
    msg = Message()
    msg["Foo"] = "bar"
    msg["Baz"] = "qux"

    h = Headers()
    h.extend(msg)
    assert h["Foo"] == "bar"
    assert h["Baz"] == "qux"



from http.cookiejar import CookieJar

import pytest

from geventhttpclient.header import Headers
from geventhttpclient.useragent import BadStatusCode, UserAgent
from tests.common import HTTPBIN_HOST, LISTENER_URL, check_upload, wsgiserver


@pytest.fixture
def tmp_file(tmp_path):
    fpath = tmp_path / "tmp.bin"
    with open(fpath, "wb") as f:
        f.write(b"123456789")
    return fpath


def internal_server_error():
    def wsgi_handler(env, start_response):
        start_response("500 Internal Server Error", [])
        return []

    return wsgi_handler


def check_redirect():
    def wsgi_handler(env, start_response):
        path_info = env.get("PATH_INFO")
        if path_info == "/":
            start_response("301 Moved Permanently", [("Location", LISTENER_URL + "redirected")])
            return []
        else:
            assert path_info == "/redirected"
            start_response("200 OK", [])
            return [b"redirected"]

    return wsgi_handler


def check_querystring():
    def wsgi_handler(env, start_response):
        querystring = env["QUERY_STRING"]
        start_response("200 OK", [("Content-type", "text/plaim")])
        return [querystring.encode("utf-8")]

    return wsgi_handler


def set_cookie():
    def wsgi_handler(env, start_response):
        start_response("200 OK", [("Set-Cookie", "testcookie=testdata")])
        return []

    return wsgi_handler


def set_cookie_401():
    def wsgi_handler(env, start_response):
        start_response("401 Unauthorized", [("Set-Cookie", "testcookie=testdata")])
        return []

    return wsgi_handler


def return_brotli():
    def wsgi_handler(env, start_response):
        path_info = env.get("PATH_INFO")
        if path_info == "/":
            start_response("200 OK", [("Content-Encoding", "br")])
        return [
            b"\x1b'\x00\x98\x04rq\x88\xa1'\xbf]\x12\xac+g!%\x98\xf4\x02\xc4\xda~)8\xba\x06xO\x11)Y\x02"
        ]

    return wsgi_handler


def test_unicode_post():
    byte_string = b"\xc8\xb9\xc8\xbc\xc9\x85"
    unicode_string = byte_string.decode("utf-8")
    headers = {
        "Content-Length": str(len(byte_string)),
        "Content-Type": "text/plain; charset=utf-8",
    }
    with wsgiserver(check_upload(byte_string, headers)):
        useragent = UserAgent()
        useragent.urlopen(LISTENER_URL, method="POST", payload=unicode_string)


def test_bytes_post():
    headers = {"Content-Length": "5", "Content-Type": "application/octet-stream"}
    with wsgiserver(check_upload(b"12345", headers)):
        useragent = UserAgent()
        useragent.urlopen(LISTENER_URL, method="POST", payload=b"12345")


def test_dict_post_with_content_type():
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    payload = {"foo": "bar"}
    with wsgiserver(set_cookie()):  # lazy. I just want to see that we dont crash making the request
        resp = UserAgent().urlopen(LISTENER_URL, method="POST", payload=payload, headers=headers)
        assert resp.status_code == 200


def test_file_post(tmp_file):
    headers = {"Content-Length": "9", "Content-Type": "application/octet-stream"}
    with wsgiserver(check_upload(b"123456789", headers)):
        useragent = UserAgent()
        with open(tmp_file, "rb") as body:
            useragent.urlopen(LISTENER_URL, method="POST", payload=body)


def test_multipart_file(tmp_file):
    with open(tmp_file, "rb") as f:
        headers = {
            "Content-Length": "173",
            "Content-Type": "multipart/form-data; boundary=custom_boundary",
        }
        files = {
            "file": (
                "report.xls",
                f,
                "application/vnd.ms-excel",
                {"Expires": "0"},
                "custom_boundary",
            )
        }

        with wsgiserver(
            check_upload(
                (
                    b"--custom_boundary\r\n"
                    b'Content-Disposition: form-data; name="file"; filename="report.xls"\r\n'
                    b"Content-Type: application/vnd.ms-excel\r\n"
                    b"Expires: 0\r\n"
                    b"\r\n"
                    b"123456789\r\n"
                    b"--custom_boundary--"
                    b"\r\n"
                ),
                headers,
            )
        ):
            useragent = UserAgent()
            useragent.urlopen(LISTENER_URL, method="POST", files=files)


def test_multipart_mixed(tmp_file):
    with open(tmp_file, "rb") as f:
        headers = {
            "Content-Length": "248",
            "Content-Type": "multipart/form-data; boundary=custom_boundary",
        }
        files = {
            "file": (
                "report.xls",
                f,
                "application/vnd.ms-excel",
                {"Expires": "0"},
                "custom_boundary",
            )
        }

        with wsgiserver(
            check_upload(
                (
                    b"--custom_boundary\r\n"
                    b'Content-Disposition: form-data; name="bla"\r\n'
                    b"\r\n"
                    b"sometext\r\n"
                    b"--custom_boundary\r\n"
                    b'Content-Disposition: form-data; name="file"; filename="report.xls"\r\n'
                    b"Content-Type: application/vnd.ms-excel\r\n"
                    b"Expires: 0\r\n"
                    b"\r\n"
                    b"123456789\r\n"
                    b"--custom_boundary--"
                    b"\r\n"
                ),
                headers,
            )
        ):
            useragent = UserAgent()
            useragent.urlopen(LISTENER_URL, method="POST", files=files, bla="sometext")


def test_redirect():
    with wsgiserver(check_redirect()):
        resp = UserAgent().urlopen(LISTENER_URL)
        assert resp.status_code == 200
        assert b"redirected" == resp.content


def test_params():
    with wsgiserver(check_querystring()):
        resp = UserAgent().urlopen(LISTENER_URL + "?param1=b", params={"param2": "hello"})
        assert resp.status_code == 200
        assert resp.content == b"param1=b&param2=hello"


def test_params_quoted():
    with wsgiserver(check_querystring()):
        resp = UserAgent().urlopen(LISTENER_URL + "?a/b", params={"path": "/"})
        assert resp.status_code == 200
        assert resp.content == b"a/b&path=%2F"


def test_server_error_with_bytes():
    with wsgiserver(internal_server_error()):
        useragent = UserAgent()
        with pytest.raises(BadStatusCode):
            useragent.urlopen(LISTENER_URL, method="POST", payload=b"12345")


def test_server_error_with_unicode():
    with wsgiserver(internal_server_error()):
        useragent = UserAgent()
        with pytest.raises(BadStatusCode):
            useragent.urlopen(LISTENER_URL, method="POST", payload="12345")


def test_server_error_with_file(tmp_file):
    with wsgiserver(internal_server_error()):
        useragent = UserAgent()
        with pytest.raises(BadStatusCode):
            with open(tmp_file, "rb") as body:
                useragent.urlopen(LISTENER_URL, method="POST", payload=body)


def test_cookiejar():
    with wsgiserver(set_cookie()):
        useragent = UserAgent(cookiejar=CookieJar())
        assert b"" == useragent.urlopen(LISTENER_URL).read()


def test_cookiejar_response_error():
    with wsgiserver(set_cookie_401()):
        useragent = UserAgent(cookiejar=CookieJar())
        with pytest.raises(BadStatusCode):
            assert b"" == useragent.urlopen(LISTENER_URL)

        assert (
            next(cookie for cookie in useragent.cookiejar if cookie.name == "testcookie").value
            == "testdata"
        )


def test_brotli_response():
    with wsgiserver(return_brotli()):
        resp = UserAgent().urlopen(LISTENER_URL, params={"path": "/"})
        assert resp.status_code == 200
        assert resp.content == b"https://github.com/gwik/geventhttpclient"


@pytest.mark.network
def test_no_form_encoded_header():
    url = f"https://{HTTPBIN_HOST}/headers"
    hdrs = Headers(UserAgent().urlopen(url).json()["headers"])
    print(hdrs)
    assert "content-type" not in hdrs
    assert "content-length" not in hdrs


@pytest.mark.network
def test_download(tmp_path):
    url = "https://proof.ovh.net/files/1Mb.dat"
    fpath = tmp_path / url.rsplit("/", 1)[-1]
    UserAgent().download(url, fpath)
    assert fpath.stat().st_size == 2**20  # 1 MB


@pytest.mark.network
def test_httpbin_multipart():
    """Sent a request body with mixed form data:

    --custom_boundary_12345
    Content-Disposition: form-data; name="bla"

    sometext
    --custom_boundary_12345
    Content-Disposition: form-data; name="file"; filename="report.xls"
    Content-Type: application/vnd.ms-excel
    Expires: 0

    1234567890
    --custom_boundary_12345--
    """

    custom_boundary = "custom_boundary_12345"
    files = {
        "file": (
            "report.xls",
            b"1234567890",
            "application/vnd.ms-excel",
            {"Expires": "0"},
            custom_boundary,
        )
    }
    resp = UserAgent().urlopen(
        f"http://{HTTPBIN_HOST}/post", method="POST", files=files, bla="sometext"
    )
    rjson = resp.json()
    request_lines = rjson["data"].splitlines()
    assert request_lines[0] == f"--{custom_boundary}"
    assert request_lines[-1] == f"--{custom_boundary}--"
    assert rjson["headers"]["Content-Type"] == [f"multipart/form-data; boundary={custom_boundary}"]
    assert rjson["files"]["file"] == ["1234567890"]
    assert rjson["form"]["bla"] == ["sometext"]



import os
import ssl
from unittest.mock import patch, MagicMock
from contextlib import contextmanager
from ssl import CertificateError
from unittest import mock

import dpkt.ssl
import gevent.queue
import gevent.server
import gevent.socket
import gevent.ssl
import pytest
from gevent import joinall
from gevent.socket import error as socket_error

from geventhttpclient import HTTPClient
from tests.common import LISTENER

BASEDIR = os.path.dirname(__file__)
KEY = os.path.join(BASEDIR, "server.key")
CERT = os.path.join(BASEDIR, "server.crt")


@contextmanager
def sslserver(handler, backlog=1):
    exception_queue = gevent.queue.Queue()

    def wrapped_handler(env, start_response):
        try:
            return handler(env, start_response)
        except Exception as e:
            exception_queue.put(e)
            raise

    server = gevent.server.StreamServer(
        LISTENER,
        backlog=backlog,
        handle=wrapped_handler,
        keyfile=KEY,
        certfile=CERT,
        ssl_version=ssl.PROTOCOL_TLS_SERVER,
    )
    server.start()
    try:
        yield server.server_host, server.server_port
        if not exception_queue.empty():
            raise exception_queue.get()
    finally:
        server.stop()


@contextmanager
def timeout_connect_server():
    sock = gevent.socket.socket(gevent.socket.AF_INET, gevent.socket.SOCK_STREAM, 0)
    sock = gevent.ssl.wrap_socket(
        sock, keyfile=KEY, certfile=CERT, ssl_version=ssl.PROTOCOL_TLS_SERVER
    )
    sock.setsockopt(gevent.socket.SOL_SOCKET, gevent.socket.SO_REUSEADDR, 1)
    sock.bind(("localhost", 0))
    sock.listen(1)

    def run(sock):
        conns = []
        while True:
            conn, addr = sock.accept()
            conns.append(conns)
            conn.recv(1024)
            gevent.sleep(10)

    job = gevent.spawn(run, sock)
    try:
        yield sock.getsockname()
        sock.close()
    finally:
        job.kill()


def simple_ssl_response(sock, addr):
    sock.recv(1024)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nConnection: close\r\n\r\n")
    sock.close()


def test_simple_ssl():
    with sslserver(simple_ssl_response) as listener:
        client = HTTPClient(*listener, insecure=True, ssl=True, ssl_options={"ca_certs": CERT})
        response = client.get("/")
        assert response.status_code == 200
        response.read()


def timeout_on_connect(sock, addr):
    sock.recv(1024)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nContent-Length: 0\r\n\r\n")


def test_implicit_sni_from_host_in_ssl():
    server_host, server_port, sent_sni = _get_sni_sent_from_client()
    assert sent_sni == server_host


def test_implicit_sni_from_header_in_ssl():
    server_host, server_port, sent_sni = _get_sni_sent_from_client(
        headers={"host": "ololo_special_host"},
    )
    assert sent_sni == "ololo_special_host"


def test_explicit_sni_in_ssl():
    server_host, server_port, sent_sni = _get_sni_sent_from_client(
        ssl_options={"server_hostname": "test_sni"},
        headers={"host": "ololo_special_host"},
    )
    assert sent_sni == "test_sni"


def _get_sni_sent_from_client(**additional_client_args):
    with sni_checker_server() as ctx:
        server_sock, server_greenlet = ctx
        server_addr, server_port = server_sock.getsockname()[:2]

        mock_addrinfo = (
            gevent.socket.AF_INET,
            gevent.socket.SOCK_STREAM,
            gevent.socket.IPPROTO_TCP,
            "localhost",
            ("127.0.0.1", server_port),
        )
        with mock.patch("gevent.socket.getaddrinfo", mock.Mock(return_value=[mock_addrinfo])):
            server_host = "some_foo"
            client = HTTPClient(
                server_host,
                server_port,
                insecure=True,
                ssl=True,
                connection_timeout=0.1,
                ssl_context_factory=gevent.ssl.create_default_context,
                **additional_client_args,
            )

            def run(http):
                try:
                    http.get("/")
                except socket_error:
                    pass  # handshake will not be completed

            client_greenlet = gevent.spawn(run, client)
            joinall([client_greenlet, server_greenlet])

    return server_host, server_port, server_greenlet.value


@contextmanager
def sni_checker_server():
    sock = gevent.socket.socket(gevent.socket.AF_INET, gevent.socket.SOCK_STREAM, 0)
    sock.setsockopt(gevent.socket.SOL_SOCKET, gevent.socket.SO_REUSEADDR, 1)
    sock.bind(("localhost", 0))
    sock.listen(1)

    # @cyberw 2021-07-10: seems this doesn't exist any more, hope it doesn't make any difference
    # sock.last_seen_sni = None

    def run(sock):
        while True:
            conn, addr = sock.accept()
            client_hello = conn.recv(4096)
            return extract_sni_from_client_hello(client_hello)

    def extract_sni_from_client_hello(hello_packet):
        records, bytes_used = dpkt.ssl.tls_multi_factory(hello_packet)

        for record in records:
            # TLS handshake only
            if record.type != 22:
                continue

            if len(record.data) == 0:
                continue
            # Client Hello only
            if record.data[0] not in (1, chr(1)):
                continue

            handshake = dpkt.ssl.TLSHandshake(record.data)

            ch = handshake.data

            SNI_extension = [
                ext_data
                for (ext_type, ext_data) in ch.extensions
                if ext_type == 0x0  # server_name
            ]
            if SNI_extension:
                SNI_extension = SNI_extension[0]
                sni_list, _ = dpkt.ssl.parse_variable_array(SNI_extension, 2)
                sni_list = sni_list[1:]  # skip SNI entry type
                first_entry, _ = dpkt.ssl.parse_variable_array(sni_list, 2)

                return first_entry.decode()

    job = gevent.spawn(run, sock)
    try:
        yield sock, job
        if job.exception:
            raise job.exception
        sock.close()
    finally:
        job.kill()


def test_timeout_on_connect():
    with timeout_connect_server() as listener:
        client = HTTPClient(*listener, insecure=True, ssl=True, ssl_options={"ca_certs": CERT})

        def run(http, wait_time=100):
            try:
                response = http.get("/")
                gevent.sleep(wait_time)
                response.read()
            except Exception:
                pass

        gevent.spawn(run, client)
        gevent.sleep(0)

        e = None
        try:
            http2 = HTTPClient(
                *listener,
                insecure=True,
                ssl=True,
                connection_timeout=0.1,
                ssl_options={"ca_certs": CERT},
            )
            http2.get("/")
        except gevent.ssl.SSLError as error:
            e = error
        except gevent.socket.timeout as error:
            e = error
        except:
            raise

        assert e is not None, "should have raised"
        if isinstance(e, gevent.ssl.SSLError):
            assert "operation timed out" in str(e)


def network_timeout(sock, addr):
    sock.recv(1024)
    gevent.sleep(10)
    sock.sendall(b"HTTP/1.1 200 Ok\r\nContent-Length: 0\r\n\r\n")


def test_network_timeout():
    with sslserver(network_timeout) as listener:
        client = HTTPClient(
            *listener,
            ssl=True,
            insecure=True,
            network_timeout=0.1,
            ssl_options={"ca_certs": CERT},
        )
        with pytest.raises(gevent.socket.timeout):
            client.get("/")


def check_client_cert_required(client):
    """Make sure hostnames are checked by default."""
    ssl_context = client._connection_pool.ssl_context
    assert ssl_context.check_hostname
    assert ssl_context.verify_mode == gevent.ssl.CERT_REQUIRED
    for socket in client._connection_pool._socket_queue.queue:
        assert socket._context.verify_mode == gevent.ssl.CERT_REQUIRED


def test_verify_self_signed_fail(capsys):
    with sslserver(simple_ssl_response) as listener:
        client = HTTPClient(*listener, ssl=True)
        with pytest.raises(CertificateError) as raised:
            client.get("/")
        assert "CERTIFICATE_VERIFY_FAILED" in str(raised.value)
        assert raised.value.verify_message == "self-signed certificate"
        check_client_cert_required(client)
        client.close()

    # This tests breaking server side socket confusingly prints its certificate error message delayed
    # into other tests output, if we don't give it a split second for printing now.
    gevent.sleep(0.01)
    captured = capsys.readouterr().err
    assert "ssl.SSLError" in captured
    assert "ALERT_UNKNOWN_CA" in captured


@patch("ssl.create_default_context")
def test_ssl_context_cert_and_keyfile(mock_create_default_context):
    mock_ssl_context = MagicMock()
    mock_create_default_context.return_value = mock_ssl_context

    ssl_options = {
        "certfile": "/path/to/certfile.pem",
        "keyfile": "/path/to/keyfile.pem",
        "ca_certs": "/path/to/ca-certificates.crt",
    }
    http_client = HTTPClient(
        "github.com", ssl_context_factory=ssl.create_default_context, ssl_options=ssl_options
    )

    mock_create_default_context.assert_called_once_with(cafile=ssl_options["ca_certs"])
    mock_ssl_context.load_cert_chain.assert_called_once_with(
        certfile=ssl_options["certfile"], keyfile=ssl_options["keyfile"]
    )
    assert isinstance(http_client, HTTPClient)


@pytest.mark.network
def test_client_ssl():
    client = HTTPClient("github.com", ssl=True)
    assert client.port == 443
    response = client.get("/")
    assert response.status_code == 200
    body = response.read()
    assert len(body)
    check_client_cert_required(client)


@pytest.mark.network
def test_fail_invalid_ca_certificate():
    certs = os.path.join(os.path.dirname(os.path.abspath(__file__)), "oncert.pem")
    client = HTTPClient("github.com", ssl_options={"ca_certs": certs})
    assert client.port == 443
    with pytest.raises(gevent.ssl.SSLError) as e_info:
        client.get("/")
    assert e_info.value.reason == "CERTIFICATE_VERIFY_FAILED"
    check_client_cert_required(client)



from distutils.core import setup

from setuptools.extension import Extension

http_parser = Extension(
    "geventhttpclient._parser",
    sources=[
        "ext/_parser.c",
        "llhttp/src/api.c",
        "llhttp/src/http.c",
        "llhttp/src/llhttp.c",
    ],
    include_dirs=[
        "ext",
        "llhttp/include",
    ],
)

setup(
    ext_modules=[http_parser],
)



from collections.abc import Mapping, MutableMapping

_dict_setitem = dict.__setitem__
_dict_getitem = dict.__getitem__
_dict_delitem = dict.__delitem__
_dict_contains = dict.__contains__
_dict_setdefault = dict.setdefault
_dict_items = dict.items
_dict_values = dict.values


class Headers(dict):
    """
    :param headers:
        An iterable of field-value pairs.

    :param kwargs:
        Additional field-value pairs to pass in to ``Headers.extend``.

    A ``dict`` like container for storing HTTP Headers.

    According to RFC 7230, header field names have to be treated case-insentitive.
    This means `SET-COOKIE` as header field denotes the same field as `set-cookie`.

    In order to implement this behavior efficiently, all header names are lowered
    and used as dictionary key for fast access of certain header lines. The original
    case-sensitive header field is stored alongside the values under the hood.

    Iterating over Header.items() returns all field-value pairs with the original
    cases. The order of the returned items matches the input order, except that
    headers with a matching header field are grouped.

    Use Header.__setitem__() and Header.update() for overwriting the content
    of matching header fields and .add() and .extend() for appending header
    lines instead of overwriting them.

    Note: b"asdf" and "asdf" are separate things. This class tries not to
    enforce the one or the other.

    >>> headers = Headers()
    >>> headers.add('Set-Cookie', 'foo=bar')
    >>> headers.add('set-cookie', 'baz=quxx')
    >>> headers['SET-cookie']
    ['foo=bar', 'baz=quxx']
    >>> headers['content-length'] = '7'
    >>> headers['Content-Length']
    '7'
    >>> headers
    Headers({'Set-Cookie': 'foo=bar, baz=quxx', 'content-length': '7'})
    """

    def __init__(self, headers=None, **kwargs):
        dict.__init__(self)
        if headers is not None:
            if isinstance(headers, type(self)):
                self._copy_from(headers)
            else:
                self.extend(headers)
        if kwargs:
            self.extend(kwargs)

    def __setitem__(self, field, value):
        return _dict_setitem(self, field.lower(), (field, value))

    def __getitem__(self, field):
        vals = _dict_getitem(self, field.lower())
        if isinstance(vals, tuple):
            return vals[1]
        return [val[1] for val in vals]

    def __delitem__(self, field):
        return _dict_delitem(self, field.lower())

    def __contains__(self, field):
        return _dict_contains(self, field.lower())

    def __eq__(self, other):
        if not isinstance(other, Mapping) and not hasattr(other, "keys"):
            return False
        if not isinstance(other, type(self)):
            other = type(self)(other)
        return {f1: self[f1] for f1 in self} == {f2: other[f2] for f2 in other}

    def __ne__(self, other):
        return not self.__eq__(other)

    values = MutableMapping.values
    get = MutableMapping.get
    keys = MutableMapping.keys

    __marker = object()

    def pop(self, field, default=__marker):
        """D.pop(field[,default]) -> value, remove specified field and return the corresponding value.
        If field is not found, d is returned if given, otherwise KeyError is raised.
        """
        # Using the MutableMapping function directly fails due to the private marker.
        # Using ordinary dict.pop would expose the internal structures.
        # So let's reinvent the wheel.
        try:
            value = self[field]
        except KeyError:
            if default is self.__marker:
                raise
            return default
        del self[field]
        return value

    def discard(self, field):
        try:
            del self[field]
        except KeyError:
            pass

    def add(self, field, value):
        """Add a (field, value) pair without overwriting the value if it already
        exists.

        >>> headers = Headers(foo='bar')
        >>> headers.add('Foo', 'baz')
        >>> headers['foo']
        'bar, baz'
        """
        field_lower = field.lower()
        new_vals = field, value
        # Keep the common case aka no item present as fast as possible
        vals = _dict_setdefault(self, field_lower, new_vals)
        if new_vals is not vals:
            # new_vals was not inserted, as there was a previous one
            if isinstance(vals, tuple):
                # Only one item so far, create a list and append new_vals
                _dict_setitem(self, field_lower, [vals, new_vals])
            elif isinstance(vals, list):
                # Already several items got inserted, we have a list
                vals.append(new_vals)
            else:
                raise TypeError("invalid vals stored")

    def extend(self, *args, **kwargs):
        """Generic import function for any type of header-like object.
        Adapted version of MutableMapping.update in order to insert items
        with self.add instead of self.__setitem__
        """
        if len(args) > 1:
            raise TypeError(f"extend() takes at most 1 positional argument ({len(args)} given)")
        other = args[0] if len(args) >= 1 else ()

        if isinstance(other, type(self)):
            for field, value in other.items():
                self.add(field, value)
        elif isinstance(other, Mapping):
            for field in other:
                self.add(field, other[field])
        elif hasattr(other, "keys"):
            for field in other.keys():
                self.add(field, other[field])
        else:
            for field, value in other:
                self.add(field, value)

        for field, value in kwargs.items():
            self.add(field, value)

    def update(self, *args, **kwargs):
        """Generic import function for any type of header-like object.
        Adapted version of MutableMapping.update in order to overwrite items
        while preserving case-sensitive header fields.
        """
        if len(args) > 1:
            raise TypeError(f"update() takes at most 1 positional argument ({len(args)} given)")
        other = args[0] if len(args) >= 1 else ()

        if isinstance(other, type(self)):
            for field, value in other.items():
                self[field] = value
        elif isinstance(other, Mapping):
            for field in other:
                self[field] = other[field]
        elif hasattr(other, "keys"):
            for field in other.keys():
                self[field] = other[field]
        else:
            for field, value in other:
                self[field] = value

        for field, value in kwargs.items():
            self[field] = value

    def getlist(self, field):
        """Returns a list of all the values for the named field. Returns an
        empty list if the field doesn't exist.
        """
        try:
            vals = _dict_getitem(self, field.lower())
        except KeyError:
            return []
        if isinstance(vals, tuple):
            return [vals[1]]
        return [val[1] for val in vals]

    def get_all(self, field, failobj=None):
        values = self.getlist(field)
        if not values:
            return failobj
        return values

    def _copy_from(self, other):
        for field in other:
            vals = _dict_getitem(other, field)
            if isinstance(vals, list):
                # No need to clone immutable tuples, only lists
                vals = list(vals)
            _dict_setitem(self, field, vals)

    def copy(self):
        clone = type(self)()
        clone._copy_from(self)
        return clone

    def __len__(self):
        return sum(1 if isinstance(vals, tuple) else len(vals) for vals in _dict_values(self))

    def __repr__(self):
        return f"{type(self).__name__}({dict(self.itermerged())})"

    def __str__(self):
        """Return a similar string as the original HTTP text received for parsing. Lines with
        matching header fields are grouped.
        """
        return "\n".join(f"{field}: {value}" for field, value in self.items())

    def itermerged(self):
        """Iterate over all headers, merging lines with duplicate header
        fields together into one item. The case of the first header field
        is preserved.
        """
        for field, vals in _dict_items(self):
            if isinstance(vals, tuple):
                yield vals
            else:
                # this should preserve either binary or string type
                sep = ", " if isinstance(vals[0][0], str) else b", "
                yield vals[0][0], sep.join(val[1] for val in vals)

    def compatible_dict(self):
        """Create a dictionary. Header lines with duplicate field names are
        merged into one line. This can be used for exchange with other
        libraries.
        """
        return dict(self.itermerged())

    def iterlower(self):
        """Iterate over all header lines, including duplicate ones.
        The header fields are all lowered.
        """
        for field, vals in _dict_items(self):
            if isinstance(vals, tuple):
                yield field, vals[1]
            else:
                for val in vals:
                    yield field, val[1]

    def items(self):
        """Iterate over all header lines, including duplicate ones."""
        for field, vals in _dict_items(self):
            if isinstance(vals, tuple):
                yield vals
            else:
                yield from vals

    # Compatibility with http.client
    getheaders = getlist
    getallmatchingheaders = getlist

    def iteroriginal(self):
        import warnings

        warnings.warn(
            "This is deprecated and will be removed in version v2.3.0. "
            "Use Headers.items() instead.",
            DeprecationWarning,
            2,
        )
        return self.items()

    def iget(self, field):
        import warnings

        warnings.warn(
            "This is deprecated and will be removed in version v2.3.0. "
            "Use Headers.getlist() instead.",
            DeprecationWarning,
            2,
        )
        return self.getlist(field)



import sys

import gevent  # noqa
import gevent.ssl  # noqa
import pytest


class DisableSSL:
    def __enter__(self):
        self._modules = dict()
        # pretend there is no ssl support
        self._modules["ssl"] = sys.modules.pop("ssl", None)
        sys.modules["ssl"] = None

        # ensure gevent must be re-imported to fire an ssl ImportError
        for module_name in [k for k in sys.modules.keys() if k.startswith("gevent")]:
            self._modules[module_name] = sys.modules.pop(module_name)

    def __exit__(self, *args, **kwargs):
        # Restore all previously disabled modules
        sys.modules.update(self._modules)


def test_import_with_nossl():
    return
    with DisableSSL():
        from geventhttpclient import HTTPClient, httplib


def test_httpclient_raises_with_no_ssl():
    return
    with DisableSSL():
        from geventhttpclient import HTTPClient

        with pytest.raises(Exception):
            HTTPClient.from_url("https://somesslhost.org/")



import http.client
import urllib.request

import pytest

from geventhttpclient.httplib import HTTPConnection, patched
from tests.common import HTTPBIN_HOST, LISTENER, server


def wrong_response_status_line(sock, addr):
    sock.recv(4096)
    sock.sendall(b"HTTP/1.1 apfais df0 asdf\r\n\r\n")


def test_httplib_exception():
    with server(wrong_response_status_line):
        connection = HTTPConnection(*LISTENER)
        connection.request("GET", "/")
        with pytest.raises(http.client.HTTPException):
            connection.getresponse()


def success_response(sock, addr):
    sock.recv(4096)
    sock.sendall(
        b"HTTP/1.1 200 Ok\r\n"
        b"Content-Type: text/plain\r\n"
        b"Set-Cookie: foo=bar\r\n"
        b"Set-Cookie: baz=bar\r\n"
        b"Content-Length: 12\r\n\r\n"
        b"Hello World!"
    )


def test_success_response():
    with server(success_response):
        connection = HTTPConnection(*LISTENER)
        connection.request("GET", "/")
        response = connection.getresponse()
        assert response.should_keep_alive()
        assert response.message_complete
        assert not response.should_close()
        assert response.read().decode() == "Hello World!"
        assert response.content_length == 12


def test_msg():
    with server(success_response):
        connection = HTTPConnection(*LISTENER)
        connection.request("GET", "/")
        response = connection.getresponse()

        assert response.msg["Set-Cookie"] == "foo=bar, baz=bar"
        assert response.msg["Content-Type"] == "text/plain"


def test_patched():
    assert http.client.HTTPResponse.__module__ == "http.client"
    assert http.client.HTTPConnection.__module__ == "http.client"
    with patched():
        assert http.client.HTTPResponse.__module__ == "geventhttpclient.httplib"
        assert http.client.HTTPConnection.__module__ == "geventhttpclient.httplib"
    assert http.client.HTTPResponse.__module__ == "http.client"
    assert http.client.HTTPConnection.__module__ == "http.client"


@pytest.mark.network
@pytest.mark.parametrize("url", [f"http://{HTTPBIN_HOST}", "https://github.com"])
def test_urllib_request(url):
    with patched():
        content = urllib.request.urlopen(url).read()
        assert content
        assert b"body" in content
